@@ -76,6 +76,8 @@
 
 #include <sys/syslog.h>
 
+int atalk_detach(caddr_t  cookie);
+
 /*
  * Setup the filter (passed in) for the requested AppleTalk values.
  *  Verifying that it's OK to do this is done by the caller once
@@ -114,7 +116,7 @@ enable_atalk(struct BlueFilter *bf, void *data, struct blueCtlBlock *ifb)
      *  Can't just hang on to the 'ifa' or sockaddr pointer,
      *  because that might get thrown away in an update.
      */
-    ifp = sotondrvcb(ifb->ifb_so)->nd_if;
+    ifp = ndrv_get_ifp(ifb->ifb_so->so_pcb);
     bzero((caddr_t)&ifb->XAtalkAddr, sizeof (ifb->XAtalkAddr));
     
     /* Assume the first one is it, since AppleTalk only uses one */
@@ -351,7 +353,14 @@ atalk_attach_protofltr(struct ifnet *ifp, struct blueCtlBlock *ifb)
 {
     u_long at_dltag, aarp_dltag;
     int retval=0;
-    struct dlil_pr_flt_str atalk_pfilter = { (caddr_t)ifb, atalk_infltr, atalk_outfltr, 0,0,0};
+    struct dlil_pr_flt_str atalk_pfilter =
+            { (caddr_t)ifb,
+              atalk_infltr,
+              atalk_outfltr,
+              0,
+              0,
+              atalk_detach
+            };
 
     /* Note: this assume the folowing here:
 	- if AppleTalk X is already up and running, get it's home port and
@@ -424,7 +433,7 @@ si_send_eth_atalk(register struct mbuf **m_orig, struct blueCtlBlock *ifb)
                 p1 = mtod(m1, unsigned char *);   /* Point to destination media addr */
                 MDATA_ETHER_END(m1);
                 /* find the interface of the X AppleTalk stack to inject this packet on? */
-                m1->m_pkthdr.rcvif = sotondrvcb(ifb->ifb_so)->nd_if;
+                m1->m_pkthdr.rcvif = ndrv_get_ifp(ifb->ifb_so->so_pcb);
                 dlil_inject_pr_input(m1, (char *)p1, ifb->atalk_proto_filter_id);
             } else
                 ifb->no_bufs2++;
@@ -448,7 +457,7 @@ si_send_eth_atalk(register struct mbuf **m_orig, struct blueCtlBlock *ifb)
 #endif
                         MDATA_ETHER_END(m);
                         /* send this packet to the X AT stack, not the network */
-                        m->m_pkthdr.rcvif = sotondrvcb(ifb->ifb_so)->nd_if;
+		                m->m_pkthdr.rcvif = ndrv_get_ifp(ifb->ifb_so->so_pcb);
                         dlil_inject_pr_input(m, p, ifb->atalk_proto_filter_id);
                         return(EJUSTRETURN);
                     } else
@@ -465,7 +474,7 @@ si_send_eth_atalk(register struct mbuf **m_orig, struct blueCtlBlock *ifb)
                     {
                         p1 = mtod(m1, unsigned char *);   /* Point to destination media addr */
                         MDATA_ETHER_END(m1);
-                        m1->m_pkthdr.rcvif = sotondrvcb(ifb->ifb_so)->nd_if;
+                        m1->m_pkthdr.rcvif = ndrv_get_ifp(ifb->ifb_so->so_pcb);
                         dlil_inject_pr_input(m1, p1, ifb->atalk_proto_filter_id);
                     } else
                         ifb->no_bufs2++;
@@ -494,14 +503,36 @@ atalk_stop(struct blueCtlBlock *ifb)
 
     if (ifb == NULL)
         return(0);
+    
+    ifb->atalk_stopping = 1;
     if (ifb->atalk_proto_filter_id) {
 #if SIP_DEBUG
         log(LOG_WARNING, "atalk_stop: deregister AppleTalk proto filter tag=%d\n",
             ifb->atalk_proto_filter_id);
 #endif
         retval = dlil_detach_filter(ifb->atalk_proto_filter_id);
-        if (retval == 0)
-            ifb->atalk_proto_filter_id = 0; /* make sure we can't deregister twice */
      }
+     
+     ifb->atalk_stopping = 0;
      return(retval);
 }
+
+/* Handle our filter being detached */
+int atalk_detach(caddr_t  cookie)
+{
+    /* Assume the interface has been detached */
+    struct blueCtlBlock *ifb = (struct blueCtlBlock*)cookie;
+    
+    ifb->atalk_proto_filter_id = 0;
+    
+    if (!ifb->atalk_stopping)
+    {
+        /*
+         * we're being detached outside the context of 
+         * atalk_stop.
+         */
+        atalk_stop(ifb);
+    }
+    
+    return 0;
+}
